package eu.e99.svc.server;

import eu.e99.svc.common.mojang.GameProfile;
import eu.e99.svc.common.mojang.MojangAPI;
import eu.e99.svc.common.protocol.*;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

public class Server {

    private final ServerSocket socket;
    private final Map<UUID, Client> clients;

    public Server(ServerSocket socket) {
        this.socket = socket;
        this.clients = new ConcurrentHashMap<>();
    }

    public void run() {
        while (true) {
            try {
                Socket client = this.socket.accept();

                // We can somewhat safely assume that we can always spawn a new virtual thread.
                Thread.ofVirtual()
                        .name("Client Handler")
                        .start(() -> this.acceptClient(client));
            } catch (Exception e) {
                System.err.printf("Failed to accept client:%n");
                e.printStackTrace(System.err);
            }
        }
    }

    private void acceptClient(Socket client) {
        try (client) {
            System.out.printf("Accepted client %s%n", client.getInetAddress());
            GameProfile profile = this.handleClient(client);
            if (profile != null) {
                this.clients.remove(profile.uuid());
            }

            System.out.printf("Closed client %s%n", client.getInetAddress());
        } catch (Exception e) {
            System.err.printf("Failed to handle client:%n");
            e.printStackTrace(System.err);
        }
    }

    /*
    - Client connects
    - Client and server switch to HANDSHAKING phase
    - Client sends HandshakePacket
    - Server sends AuthRequestPacket
        - server_id
    - Client and server switch to AUTHENTICATING phase
    - Client sends AuthResponsePacket
        - username
    - Server sends AuthSuccessPacket or Disconnect
    - Client and server switch to SIGNALLING phase
    - Client can send SignalPacket, server distributes them to the target
    */
    private GameProfile handleClient(final Socket client) throws Exception {
        final DataInputStream in = new DataInputStream(client.getInputStream());
        final DataOutputStream out = new DataOutputStream(client.getOutputStream());

        final String serverId = MojangAPI.generateServerId();
        PacketPhase phase = PacketPhase.HANDSHAKING;
        GameProfile profile = null;

        while (true) {
            Packet packet = Packets.read(in);
            if (packet == null) {
                return profile;
            }

            switch (packet) {
                case HandshakePacket handshake -> {
                    if (phase != PacketPhase.HANDSHAKING) {
                        break;
                    }

                    if (handshake.protocolVersion > Packets.VERSION) {
                        DisconnectPacket disconnect = new DisconnectPacket();
                        disconnect.reason = "Outdated server, please downgrade your client.";
                        Packets.write(disconnect, out);

                        return profile;
                    }

                    if (handshake.protocolVersion < Packets.VERSION) {
                        DisconnectPacket disconnect = new DisconnectPacket();
                        disconnect.reason = "Outdated client, please upgrade your client.";
                        Packets.write(disconnect, out);

                        return profile;
                    }

                    AuthRequestPacket authRequest = new AuthRequestPacket();
                    authRequest.serverId = serverId;
                    Packets.write(authRequest, out);

                    phase = PacketPhase.AUTHENTICATING;
                }

                case AuthResponsePacket authResponse -> {
                    if (phase != PacketPhase.AUTHENTICATING) {
                        break;
                    }

                    profile = MojangAPI.hasJoined(authResponse.username);
                    if (profile == null) {
                        DisconnectPacket disconnect = new DisconnectPacket();
                        disconnect.reason = "Failed to authenticate using Mojang servers.";
                        Packets.write(disconnect, out);

                        return profile;
                    }

                    boolean alreadyConnected = this.clients.containsKey(profile.uuid());
                    if (alreadyConnected) {
                        DisconnectPacket disconnect = new DisconnectPacket();
                        disconnect.reason = "You are already connected.";
                        Packets.write(disconnect, out);

                        return profile;
                    }

                    Client cli = toSend -> {
                        Packets.write(toSend, out);
                    };

                    this.clients.put(profile.uuid(), cli);

                    // Inform the client that they authenticated successfully
                    AuthSuccessPacket success = new AuthSuccessPacket();
                    Packets.write(success, out);

                    phase = PacketPhase.SIGNALLING;
                }

                case SignalPacket signal -> {
                    if (phase != PacketPhase.SIGNALLING) {
                        break;
                    }

                    // Overwrite from to fix spoofing messaging.
                    signal.from = profile.uuid();
                    Client to = this.clients.get(signal.to);
                    if (to == null) {
                        // TODO: Send the client back some error code?
                        break;
                    }

                    // TODO: Do we want a maximum length here?
                    to.sendPacket(signal);
                }

                case DisconnectPacket disconnect -> {
                    System.out.printf("Client %s disconnected with reason \"%s\")%n", client.getInetAddress(), disconnect.reason);
                    return profile;
                }

                default -> {
                    System.out.printf("Received unknown packet from %s in %s phase%n", client.getInetAddress(), phase);
                }
            }
        }
    }
}
